@@ -18,7 +18,10 @@ from ipaddress import (
 from json import dumps, loads
 from uuid import UUID
 
+from dateutil.parser import ParserError, parse
+
 from pg8000.exceptions import InterfaceError
+from pg8000.types import PGInterval, Range
 
 
 ANY_ARRAY = 2277
@@ -36,11 +39,17 @@ CSTRING = 2275
 CSTRING_ARRAY = 1263
 DATE = 1082
 DATE_ARRAY = 1182
+DATEMULTIRANGE = 4535
+DATERANGE = 3912
 FLOAT = 701
 FLOAT_ARRAY = 1022
 INET = 869
 INET_ARRAY = 1041
 INT2VECTOR = 22
+INT4MULTIRANGE = 4451
+INT4RANGE = 3904
+INT8MULTIRANGE = 4536
+INT8RANGE = 3926
 INTEGER = 23
 INTEGER_ARRAY = 1007
 INTERVAL = 1186
@@ -57,11 +66,14 @@ NAME = 19
 NAME_ARRAY = 1003
 NUMERIC = 1700
 NUMERIC_ARRAY = 1231
+NUMRANGE = 3906
+NUMMULTIRANGE = 4532
 NULLTYPE = -1
 OID = 26
 POINT = 600
 REAL = 700
 REAL_ARRAY = 1021
+RECORD = 2249
 SMALLINT = 21
 SMALLINT_ARRAY = 1005
 SMALLINT_VECTOR = 22
@@ -74,6 +86,10 @@ TIMESTAMP = 1114
 TIMESTAMP_ARRAY = 1115
 TIMESTAMPTZ = 1184
 TIMESTAMPTZ_ARRAY = 1185
+TSMULTIRANGE = 4533
+TSRANGE = 3908
+TSTZMULTIRANGE = 4534
+TSTZRANGE = 3910
 UNKNOWN = 705
 UUID_TYPE = 2950
 UUID_ARRAY = 2951
@@ -115,7 +131,11 @@ def date_in(data):
     if data in ("infinity", "-infinity"):
         return data
     else:
-        return Datetime.strptime(data, "%Y-%m-%d").date()
+        try:
+            return Datetime.strptime(data, "%Y-%m-%d").date()
+        except ValueError:
+            # pg date can overflow Python Datetime
+            return data
 
 
 def date_out(v):
@@ -154,25 +174,11 @@ def int_out(v):
 
 
 def interval_in(data):
-    t = {}
-
-    curr_val = None
-    for k in data.split():
-        if ":" in k:
-            t["hours"], t["minutes"], t["seconds"] = map(float, k.split(":"))
-        else:
-            try:
-                curr_val = float(k)
-            except ValueError:
-                t[PGInterval.UNIT_MAP[k]] = curr_val
-
-    for n in ["weeks", "months", "years", "decades", "centuries", "millennia"]:
-        if n in t:
-            raise InterfaceError(
-                f"Can't fit the interval {t} into a datetime.timedelta."
-            )
-
-    return Timedelta(**t)
+    pg_interval = PGInterval.from_str(data)
+    try:
+        return pg_interval.to_timedelta()
+    except ValueError:
+        return pg_interval
 
 
 def interval_out(v):
@@ -199,6 +205,10 @@ def numeric_out(d):
     return str(d)
 
 
+def point_in(data):
+    return tuple(map(float, data[1:-1].split(",")))
+
+
 def pg_interval_in(data):
     return PGInterval.from_str(data)
 
@@ -207,6 +217,17 @@ def pg_interval_out(v):
     return str(v)
 
 
+def range_out(v):
+    if v.is_empty:
+        return "empty"
+    else:
+        le = v.lower
+        val_lower = "" if le is None else make_param(PY_TYPES, le)
+        ue = v.upper
+        val_upper = "" if ue is None else make_param(PY_TYPES, ue)
+        return f"{v.bounds[0]}{val_lower},{val_upper}{v.bounds[1]}"
+
+
 def string_in(data):
     return data
 
@@ -228,13 +249,30 @@ def timestamp_in(data):
     if data in ("infinity", "-infinity"):
         return data
 
-    pattern = "%Y-%m-%d %H:%M:%S.%f" if "." in data else "%Y-%m-%d %H:%M:%S"
-    return Datetime.strptime(data, pattern)
+    try:
+        pattern = "%Y-%m-%d %H:%M:%S.%f" if "." in data else "%Y-%m-%d %H:%M:%S"
+        return Datetime.strptime(data, pattern)
+    except ValueError:
+        try:
+            return parse(data)
+        except ParserError:
+            # pg timestamp can overflow Python Datetime
+            return data
 
 
 def timestamptz_in(data):
-    patt = "%Y-%m-%d %H:%M:%S.%f%z" if "." in data else "%Y-%m-%d %H:%M:%S%z"
-    return Datetime.strptime(data + "00", patt)
+    if data in ("infinity", "-infinity"):
+        return data
+
+    try:
+        patt = "%Y-%m-%d %H:%M:%S.%f%z" if "." in data else "%Y-%m-%d %H:%M:%S%z"
+        return Datetime.strptime(f"{data}00", patt)
+    except ValueError:
+        try:
+            return parse(data)
+        except ParserError:
+            # pg timestamptz can overflow Python Datetime
+            return data
 
 
 def unknown_out(v):
@@ -242,7 +280,7 @@ def unknown_out(v):
 
 
 def vector_in(data):
-    return eval("[" + data.replace(" ", ",") + "]")
+    return [int(v) for v in data.split()]
 
 
 def uuid_out(v):
@@ -253,146 +291,7 @@ def uuid_in(data):
     return UUID(data)
 
 
-class PGInterval:
-    UNIT_MAP = {
-        "year": "years",
-        "years": "years",
-        "millennia": "millennia",
-        "millenium": "millennia",
-        "centuries": "centuries",
-        "century": "centuries",
-        "decades": "decades",
-        "decade": "decades",
-        "years": "years",
-        "year": "years",
-        "months": "months",
-        "month": "months",
-        "mon": "months",
-        "mons": "months",
-        "weeks": "weeks",
-        "week": "weeks",
-        "days": "days",
-        "day": "days",
-        "hours": "hours",
-        "hour": "hours",
-        "minutes": "minutes",
-        "minute": "minutes",
-        "seconds": "seconds",
-        "second": "seconds",
-        "microseconds": "microseconds",
-        "microsecond": "microseconds",
-    }
-
-    @staticmethod
-    def from_str(interval_str):
-        t = {}
-
-        curr_val = None
-        for k in interval_str.split():
-            if ":" in k:
-                hours_str, minutes_str, seconds_str = k.split(":")
-                hours = int(hours_str)
-                if hours != 0:
-                    t["hours"] = hours
-                minutes = int(minutes_str)
-                if minutes != 0:
-                    t["minutes"] = minutes
-                try:
-                    seconds = int(seconds_str)
-                except ValueError:
-                    seconds = float(seconds_str)
-
-                if seconds != 0:
-                    t["seconds"] = seconds
-
-            else:
-                try:
-                    curr_val = int(k)
-                except ValueError:
-                    t[PGInterval.UNIT_MAP[k]] = curr_val
-
-        return PGInterval(**t)
-
-    def __init__(
-        self,
-        millennia=None,
-        centuries=None,
-        decades=None,
-        years=None,
-        months=None,
-        weeks=None,
-        days=None,
-        hours=None,
-        minutes=None,
-        seconds=None,
-        microseconds=None,
-    ):
-        self.millennia = millennia
-        self.centuries = centuries
-        self.decades = decades
-        self.years = years
-        self.months = months
-        self.weeks = weeks
-        self.days = days
-        self.hours = hours
-        self.minutes = minutes
-        self.seconds = seconds
-        self.microseconds = microseconds
-
-    def __repr__(self):
-        return f"<PGInterval {self}>"
-
-    def __str__(self):
-        pairs = (
-            ("millennia", self.millennia),
-            ("centuries", self.centuries),
-            ("decades", self.decades),
-            ("years", self.years),
-            ("months", self.months),
-            ("weeks", self.weeks),
-            ("days", self.days),
-            ("hours", self.hours),
-            ("minutes", self.minutes),
-            ("seconds", self.seconds),
-            ("microseconds", self.microseconds),
-        )
-        return " ".join(f"{v} {n}" for n, v in pairs if v is not None)
-
-    def normalize(self):
-        months = 0
-        if self.months is not None:
-            months += self.months
-        if self.years is not None:
-            months += self.years * 12
-
-        days = 0
-        if self.days is not None:
-            days += self.days
-        if self.weeks is not None:
-            days += self.weeks * 7
-
-        seconds = 0
-        if self.hours is not None:
-            seconds += self.hours * 60 * 60
-        if self.minutes is not None:
-            seconds += self.minutes * 60
-        if self.seconds is not None:
-            seconds += self.seconds
-        if self.microseconds is not None:
-            seconds += self.microseconds / 1000000
-
-        return PGInterval(months=months, days=days, seconds=seconds)
-
-    def __eq__(self, other):
-        if isinstance(other, PGInterval):
-            s = self.normalize()
-            o = other.normalize()
-            return s.months == o.months and s.days == o.days and s.seconds == o.seconds
-        else:
-            return False
-
-
-class ArrayState(Enum):
+class ParserState(Enum):
     InString = 1
     InEscape = 2
     InValue = 3
@@ -400,19 +299,19 @@ class ArrayState(Enum):
 
 
 def _parse_array(data, adapter):
-    state = ArrayState.Out
+    state = ParserState.Out
     stack = [[]]
     val = []
     for c in data:
-        if state == ArrayState.InValue:
+        if state == ParserState.InValue:
             if c in ("}", ","):
                 value = "".join(val)
                 stack[-1].append(None if value == "NULL" else adapter(value))
-                state = ArrayState.Out
+                state = ParserState.Out
             else:
                 val.append(c)
 
-        if state == ArrayState.Out:
+        if state == ParserState.Out:
             if c == "{":
                 a = []
                 stack[-1].append(a)
@@ -423,22 +322,22 @@ def _parse_array(data, adapter):
                 pass
             elif c == '"':
                 val = []
-                state = ArrayState.InString
+                state = ParserState.InString
             else:
                 val = [c]
-                state = ArrayState.InValue
+                state = ParserState.InValue
 
-        elif state == ArrayState.InString:
+        elif state == ParserState.InString:
             if c == '"':
                 stack[-1].append(adapter("".join(val)))
-                state = ArrayState.Out
+                state = ParserState.Out
             elif c == "\\":
-                state = ArrayState.InEscape
+                state = ParserState.InEscape
             else:
                 val.append(c)
-        elif state == ArrayState.InEscape:
+        elif state == ParserState.InEscape:
             val.append(c)
-            state = ArrayState.InString
+            state = ParserState.InString
 
     return stack[0][0]
 
@@ -479,7 +378,8 @@ def array_string_escape(v):
     if (
         len(val) == 0
         or val == "NULL"
-        or any([c in val for c in ("{", "}", ",", " ", "\\")])
+        or any(c.isspace() for c in val)
+        or any(c in val for c in ("{", "}", ",", "\\"))
     ):
         val = f'"{val}"'
     return val
@@ -488,10 +388,12 @@ def array_string_escape(v):
 def array_out(ar):
     result = []
     for v in ar:
-
-        if isinstance(v, (list, tuple)):
+        if isinstance(v, list):
             val = array_out(v)
 
+        elif isinstance(v, tuple):
+            val = f'"{composite_out(v)}"'
+
         elif v is None:
             val = "NULL"
 
@@ -509,7 +411,135 @@ def array_out(ar):
 
         result.append(val)
 
-    return "{" + ",".join(result) + "}"
+    return f'{{{",".join(result)}}}'
+
+
+def composite_out(ar):
+    result = []
+    for v in ar:
+        if isinstance(v, list):
+            val = array_out(v)
+
+        elif isinstance(v, tuple):
+            val = composite_out(v)
+
+        elif v is None:
+            val = ""
+
+        elif isinstance(v, dict):
+            val = array_string_escape(json_out(v))
+
+        elif isinstance(v, (bytes, bytearray)):
+            val = f'"\\{bytes_out(v)}"'
+
+        elif isinstance(v, str):
+            val = array_string_escape(v)
+
+        else:
+            val = make_param(PY_TYPES, v)
+
+        result.append(val)
+
+    return f'({",".join(result)})'
+
+
+def _range_in(elem_func):
+    def range_in(data):
+        if data == "empty":
+            return Range(is_empty=True)
+        else:
+            le, ue = [None if v == "" else elem_func(v) for v in data[1:-1].split(",")]
+            return Range(le, ue, bounds=f"{data[0]}{data[-1]}")
+
+    return range_in
+
+
+daterange_in = _range_in(date_in)
+int4range_in = _range_in(int)
+int8range_in = _range_in(int)
+numrange_in = _range_in(Decimal)
+
+
+def ts_in(data):
+    return timestamp_in(data[1:-1])
+
+
+def tstz_in(data):
+    return timestamptz_in(data[1:-1])
+
+
+tsrange_in = _range_in(ts_in)
+tstzrange_in = _range_in(tstz_in)
+
+
+def _multirange_in(adapter):
+    def f(data):
+        in_range = False
+        result = []
+        val = []
+        for c in data:
+            if in_range:
+                val.append(c)
+                if c in "])":
+                    value = "".join(val)
+                    val.clear()
+                    result.append(adapter(value))
+                    in_range = False
+            elif c in "[(":
+                val.append(c)
+                in_range = True
+
+        return result
+
+    return f
+
+
+datemultirange_in = _multirange_in(daterange_in)
+int4multirange_in = _multirange_in(int4range_in)
+int8multirange_in = _multirange_in(int8range_in)
+nummultirange_in = _multirange_in(numrange_in)
+tsmultirange_in = _multirange_in(tsrange_in)
+tstzmultirange_in = _multirange_in(tstzrange_in)
+
+
+def record_in(data):
+    state = ParserState.Out
+    results = []
+    val = []
+    for c in data:
+        if state == ParserState.InValue:
+            if c in (")", ","):
+                value = "".join(val)
+                val.clear()
+                results.append(None if value == "" else value)
+                state = ParserState.Out
+            else:
+                val.append(c)
+
+        if state == ParserState.Out:
+            if c in "(),":
+                pass
+            elif c == '"':
+                state = ParserState.InString
+            else:
+                val.append(c)
+                state = ParserState.InValue
+
+        elif state == ParserState.InString:
+            if c == '"':
+                results.append("".join(val))
+                val.clear()
+                state = ParserState.Out
+            elif c == "\\":
+                state = ParserState.InEscape
+            else:
+                val.append(c)
+
+        elif state == ParserState.InEscape:
+            val.append(c)
+            state = ParserState.InString
+
+    return tuple(results)
 
 
 PY_PG = {
@@ -543,6 +573,7 @@ PY_TYPES = {
     IPv4Network: inet_out,  # inet
     IPv6Network: inet_out,  # inet
     PGInterval: interval_out,  # interval
+    Range: range_out,  # range types
     Time: time_out,  # time
     Timedelta: interval_out,  # interval
     UUID: uuid_out,  # uuid
@@ -555,7 +586,7 @@ PY_TYPES = {
     str: string_out,  # unknown
     int: int_out,
     list: array_out,
-    tuple: array_out,
+    tuple: composite_out,
 }
 
 
@@ -573,10 +604,16 @@ PG_TYPES = {
     CSTRING_ARRAY: string_array_in,  # cstring[]
     DATE: date_in,  # date
     DATE_ARRAY: date_array_in,  # date[]
+    DATEMULTIRANGE: datemultirange_in,  # datemultirange
+    DATERANGE: daterange_in,  # daterange
     FLOAT: float,  # float8
     FLOAT_ARRAY: float_array_in,  # float8[]
     INET: inet_in,  # inet
     INET_ARRAY: inet_array_in,  # inet[]
+    INT4MULTIRANGE: int4multirange_in,  # int4multirange
+    INT4RANGE: int4range_in,  # int4range
+    INT8MULTIRANGE: int8multirange_in,  # int8multirange
+    INT8RANGE: int8range_in,  # int8range
     INTEGER: int,  # int4
     INTEGER_ARRAY: int_array_in,  # int4[]
     JSON: json_in,  # json
@@ -590,11 +627,15 @@ PG_TYPES = {
     NAME_ARRAY: string_array_in,  # name[]
     NUMERIC: numeric_in,  # numeric
     NUMERIC_ARRAY: numeric_array_in,  # numeric[]
+    NUMRANGE: numrange_in,  # numrange
+    NUMMULTIRANGE: nummultirange_in,  # nummultirange
     OID: int,  # oid
+    POINT: point_in,  # point
     INTERVAL: interval_in,  # interval
     INTERVAL_ARRAY: interval_array_in,  # interval[]
     REAL: float,  # float4
     REAL_ARRAY: float_array_in,  # float4[]
+    RECORD: record_in,  # record
     SMALLINT: int,  # int2
     SMALLINT_ARRAY: int_array_in,  # int2[]
     SMALLINT_VECTOR: vector_in,  # int2vector
@@ -607,6 +648,10 @@ PG_TYPES = {
     TIMESTAMP_ARRAY: timestamp_array_in,  # timestamp
     TIMESTAMPTZ: timestamptz_in,  # timestamptz
     TIMESTAMPTZ_ARRAY: timestamptz_array_in,  # timestamptz
+    TSMULTIRANGE: tsmultirange_in,  # tsmultirange
+    TSRANGE: tsrange_in,  # tsrange
+    TSTZMULTIRANGE: tstzmultirange_in,  # tstzmultirange
+    TSTZRANGE: tstzrange_in,  # tstzrange
     UNKNOWN: string_in,  # unknown
     UUID_ARRAY: uuid_array_in,  # uuid[]
     UUID_TYPE: uuid_in,  # uuid
@@ -725,14 +770,16 @@ def literal(value):
         return str(value)
     elif isinstance(value, (bytes, bytearray)):
         return f"X'{value.hex()}'"
+    elif isinstance(value, Datetime):
+        return f"'{datetime_out(value)}'"
     elif isinstance(value, Date):
         return f"'{date_out(value)}'"
     elif isinstance(value, Time):
         return f"'{time_out(value)}'"
-    elif isinstance(value, Datetime):
-        return f"'{datetime_out(value)}'"
     elif isinstance(value, Timedelta):
         return f"'{interval_out(value)}'"
+    elif isinstance(value, list):
+        return f"'{array_out(value)}'"
     else:
         val = str(value).replace("'", "''")
         return f"'{val}'"
