@@ -1,8 +1,8 @@
 import codecs
 import socket
-import struct
 from collections import defaultdict, deque
 from hashlib import md5
+from importlib.metadata import version
 from io import IOBase, TextIOBase
 from itertools import count
 from struct import Struct
@@ -18,10 +18,6 @@ from pg8000.converters import (
 )
 from pg8000.exceptions import DatabaseError, Error, InterfaceError
 
-try:
-    from importlib.metadata import version
-except ImportError:
-    from importlib_metadata import version
 
 ver = version("pg8000")
 
@@ -143,6 +139,109 @@ IN_TRANSACTION = b"T"
 IN_FAILED_TRANSACTION = b"E"
 
 
+def _flush(sock):
+    try:
+        sock.flush()
+    except OSError as e:
+        raise InterfaceError("network error") from e
+
+
+def _read(sock, size):
+    got = 0
+    buff = []
+    try:
+        while got < size:
+            block = sock.read(size - got)
+            if block == b"":
+                raise InterfaceError("network error")
+            buff.append(block)
+            got += len(block)
+    except OSError as e:
+        raise InterfaceError("network error") from e
+
+    return b"".join(buff)
+
+
+def _write(sock, d):
+    try:
+        sock.write(d)
+    except OSError as e:
+        raise InterfaceError("network error") from e
+
+
+def _make_socket(
+    unix_sock, sock, host, port, timeout, source_address, tcp_keepalive, ssl_context
+):
+    if unix_sock is not None:
+        if sock is not None:
+            raise InterfaceError("If unix_sock is provided, sock must be None")
+
+        try:
+            if not hasattr(socket, "AF_UNIX"):
+                raise InterfaceError(
+                    "attempt to connect to unix socket on unsupported platform"
+                )
+            sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
+            sock.settimeout(timeout)
+            sock.connect(unix_sock)
+            if tcp_keepalive:
+                sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
+        except socket.error as e:
+            if sock is not None:
+                sock.close()
+            raise InterfaceError("communication error") from e
+
+    elif sock is not None:
+        pass
+
+    elif host is not None:
+        try:
+            sock = socket.create_connection((host, port), timeout, source_address)
+        except socket.error as e:
+            raise InterfaceError(
+                f"Can't create a connection to host {host} and port {port} "
+                f"(timeout is {timeout} and source_address is {source_address})."
+            ) from e
+
+        if tcp_keepalive:
+            sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
+
+    else:
+        raise InterfaceError("one of host, sock or unix_sock must be provided")
+
+    channel_binding = None
+    if ssl_context is not None:
+        try:
+            import ssl
+
+            if ssl_context is True:
+                ssl_context = ssl.create_default_context()
+
+            request_ssl = getattr(ssl_context, "request_ssl", True)
+
+            if request_ssl:
+                # Int32(8) - Message length, including self.
+                # Int32(80877103) - The SSL request code.
+                sock.sendall(ii_pack(8, 80877103))
+                resp = sock.recv(1)
+                if resp != b"S":
+                    raise InterfaceError("Server refuses SSL")
+
+            sock = ssl_context.wrap_socket(sock, server_hostname=host)
+
+            if request_ssl:
+                channel_binding = scramp.make_channel_binding(
+                    "tls-server-end-point", sock
+                )
+
+        except ImportError:
+            raise InterfaceError(
+                "SSL required but ssl module not available in this python "
+                "installation."
+            )
+    return channel_binding, sock
+
+
 class CoreConnection:
     def __enter__(self):
         return self
@@ -164,6 +263,7 @@ class CoreConnection:
         tcp_keepalive=True,
         application_name=None,
         replication=None,
+        sock=None,
     ):
         self._client_encoding = "utf8"
         self._commands_with_count = (
@@ -204,99 +304,24 @@ class CoreConnection:
         else:
             self.password = password
 
-        self.autocommit = False
         self._xid = None
         self._statement_nums = set()
 
         self._caches = {}
 
-        if unix_sock is None and host is not None:
-            try:
-                self._usock = socket.create_connection(
-                    (host, port), timeout, source_address
-                )
-            except socket.error as e:
-                raise InterfaceError(
-                    f"Can't create a connection to host {host} and port {port} "
-                    f"(timeout is {timeout} and source_address is {source_address})."
-                ) from e
-
-        elif unix_sock is not None:
-            try:
-                if not hasattr(socket, "AF_UNIX"):
-                    raise InterfaceError(
-                        "attempt to connect to unix socket on unsupported platform"
-                    )
-                self._usock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
-                self._usock.settimeout(timeout)
-                self._usock.connect(unix_sock)
-            except socket.error as e:
-                if self._usock is not None:
-                    self._usock.close()
-                raise InterfaceError("communication error") from e
-
-        else:
-            raise InterfaceError("one of host or unix_sock must be provided")
-
-        if tcp_keepalive:
-            self._usock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
-
-        self.channel_binding = None
-        if ssl_context is not None:
-            try:
-                import ssl
-
-                if ssl_context is True:
-                    ssl_context = ssl.create_default_context()
-
-                request_ssl = getattr(ssl_context, "request_ssl", True)
-
-                if request_ssl:
-                    # Int32(8) - Message length, including self.
-                    # Int32(80877103) - The SSL request code.
-                    self._usock.sendall(ii_pack(8, 80877103))
-                    resp = self._usock.recv(1)
-                    if resp != b"S":
-                        raise InterfaceError("Server refuses SSL")
-
-                self._usock = ssl_context.wrap_socket(self._usock, server_hostname=host)
-
-                if request_ssl:
-                    self.channel_binding = scramp.make_channel_binding(
-                        "tls-server-end-point", self._usock
-                    )
-
-            except ImportError:
-                raise InterfaceError(
-                    "SSL required but ssl module not available in this python "
-                    "installation."
-                )
+        self.channel_binding, self._usock = _make_socket(
+            unix_sock,
+            sock,
+            host,
+            port,
+            timeout,
+            source_address,
+            tcp_keepalive,
+            ssl_context,
+        )
 
         self._sock = self._usock.makefile(mode="rwb")
 
-        def sock_flush():
-            try:
-                self._sock.flush()
-            except OSError as e:
-                raise InterfaceError("network error") from e
-
-        self._flush = sock_flush
-
-        def sock_read(b):
-            try:
-                return self._sock.read(b)
-            except OSError as e:
-                raise InterfaceError("network error") from e
-
-        self._read = sock_read
-
-        def sock_write(d):
-            try:
-                self._sock.write(d)
-            except OSError as e:
-                raise InterfaceError("network error") from e
-
-        self._write = sock_write
         self._backend_key_data = None
 
         self.pg_types = defaultdict(lambda: string_in, PG_TYPES)
@@ -337,22 +362,17 @@ class CoreConnection:
         for k, v in init_params.items():
             val.extend(k.encode("ascii") + NULL_BYTE + v + NULL_BYTE)
         val.append(0)
-        self._write(i_pack(len(val) + 4))
-        self._write(val)
-        self._flush()
+        _write(self._sock, i_pack(len(val) + 4))
+        _write(self._sock, val)
+        _flush(self._sock)
 
         try:
-
             code = None
             context = Context(None)
             while code not in (READY_FOR_QUERY, ERROR_RESPONSE):
+                code, data_len = ci_unpack(_read(self._sock, 5))
 
-                try:
-                    code, data_len = ci_unpack(self._read(5))
-                except struct.error as e:
-                    raise InterfaceError("network error") from e
-
-                self.message_types[code](self._read(data_len - 4), context)
+                self.message_types[code](_read(self._sock, data_len - 4), context)
 
             if context.error is not None:
                 raise context.error
@@ -471,10 +491,10 @@ class CoreConnection:
                 bytes_read = readinto(bffr)
                 if bytes_read == 0:
                     break
-                self._write(COPY_DATA)
-                self._write(i_pack(bytes_read + 4))
-                self._write(bffr[:bytes_read])
-                self._flush()
+                _write(self._sock, COPY_DATA)
+                _write(self._sock, i_pack(bytes_read + 4))
+                _write(self._sock, bffr[:bytes_read])
+                _flush(self._sock)
 
         else:
             for k in context.stream:
@@ -489,12 +509,12 @@ class CoreConnection:
                     b = k
 
                 self._send_message(COPY_DATA, b)
-                self._flush()
+                _flush(self._sock)
 
         # Send CopyDone
-        self._write(COPY_DONE_MSG)
-        self._write(SYNC_MSG)
-        self._flush()
+        _write(self._sock, COPY_DONE_MSG)
+        _write(self._sock, SYNC_MSG)
+        _flush(self._sock)
 
     def handle_NOTIFICATION_RESPONSE(self, data, context):
         """https://www.postgresql.org/docs/current/protocol-message-formats.html"""
@@ -517,8 +537,8 @@ class CoreConnection:
             raise InterfaceError("connection is closed")
 
         try:
-            self._write(TERMINATE_MSG)
-            self._flush()
+            _write(self._sock, TERMINATE_MSG)
+            _flush(self._sock)
         finally:
             try:
                 self._usock.close()
@@ -541,7 +561,7 @@ class CoreConnection:
                     "provided"
                 )
             self._send_message(PASSWORD, self.password + NULL_BYTE)
-            self._flush()
+            _flush(self._sock)
 
         elif auth_code == 5:
             salt = b"".join(cccc_unpack(data, 4))
@@ -554,7 +574,7 @@ class CoreConnection:
                 md5(self.password + self.user).hexdigest().encode("ascii") + salt
             ).hexdigest().encode("ascii")
             self._send_message(PASSWORD, pwd + NULL_BYTE)
-            self._flush()
+            _flush(self._sock)
 
         elif auth_code == 10:
             # AuthenticationSASL
@@ -572,7 +592,7 @@ class CoreConnection:
 
             # SASLInitialResponse
             self._send_message(PASSWORD, mech + i_pack(len(init)) + init)
-            self._flush()
+            _flush(self._sock)
 
         elif auth_code == 11:
             # AuthenticationSASLContinue
@@ -581,7 +601,7 @@ class CoreConnection:
             # SASLResponse
             msg = self.auth.get_client_final().encode("utf8")
             self._send_message(PASSWORD, msg)
-            self._flush()
+            _flush(self._sock)
 
         elif auth_code == 12:
             # AuthenticationSASLFinal
@@ -634,7 +654,6 @@ class CoreConnection:
             context.rows = []
 
     def send_PARSE(self, statement_name_bin, statement, oids=()):
-
         val = bytearray(statement_name_bin)
         val.extend(statement.encode(self._client_encoding) + NULL_BYTE)
         val.extend(h_pack(len(oids)))
@@ -642,11 +661,11 @@ class CoreConnection:
             val.extend(i_pack(0 if oid == -1 else oid))
 
         self._send_message(PARSE, val)
-        self._write(FLUSH_MSG)
+        _write(self._sock, FLUSH_MSG)
 
     def send_DESCRIBE_STATEMENT(self, statement_name_bin):
         self._send_message(DESCRIBE, STATEMENT + statement_name_bin)
-        self._write(FLUSH_MSG)
+        _write(self._sock, FLUSH_MSG)
 
     def send_QUERY(self, sql):
         self._send_message(QUERY, sql.encode(self._client_encoding) + NULL_BYTE)
@@ -655,7 +674,7 @@ class CoreConnection:
         context = Context(statement)
 
         self.send_QUERY(statement)
-        self._flush()
+        _flush(self._sock)
         self.handle_messages(context)
 
         return context
@@ -664,15 +683,15 @@ class CoreConnection:
         context = Context(statement, stream=stream)
 
         self.send_PARSE(NULL_BYTE, statement, oids)
-        self._write(SYNC_MSG)
-        self._flush()
+        _write(self._sock, SYNC_MSG)
+        _flush(self._sock)
         self.handle_messages(context)
         self.send_DESCRIBE_STATEMENT(NULL_BYTE)
 
-        self._write(SYNC_MSG)
+        _write(self._sock, SYNC_MSG)
 
         try:
-            self._flush()
+            _flush(self._sock)
         except AttributeError as e:
             if self._sock is None:
                 raise InterfaceError("connection is closed")
@@ -683,14 +702,13 @@ class CoreConnection:
         self.handle_messages(context)
         self.send_EXECUTE()
 
-        self._write(SYNC_MSG)
-        self._flush()
+        _write(self._sock, SYNC_MSG)
+        _flush(self._sock)
         self.handle_messages(context)
 
         return context
 
     def prepare_statement(self, statement, oids=None):
-
         for i in count():
             statement_name = f"pg8000_statement_{i}"
             statement_name_bin = statement_name.encode("ascii") + NULL_BYTE
@@ -700,10 +718,10 @@ class CoreConnection:
 
         self.send_PARSE(statement_name_bin, statement, oids)
         self.send_DESCRIBE_STATEMENT(statement_name_bin)
-        self._write(SYNC_MSG)
+        _write(self._sock, SYNC_MSG)
 
         try:
-            self._flush()
+            _flush(self._sock)
         except AttributeError as e:
             if self._sock is None:
                 raise InterfaceError("connection is closed")
@@ -722,16 +740,16 @@ class CoreConnection:
 
         self.send_BIND(statement_name_bin, params)
         self.send_EXECUTE()
-        self._write(SYNC_MSG)
-        self._flush()
+        _write(self._sock, SYNC_MSG)
+        _flush(self._sock)
         self.handle_messages(context)
         return context
 
     def _send_message(self, code, data):
         try:
-            self._write(code)
-            self._write(i_pack(len(data) + 4))
-            self._write(data)
+            _write(self._sock, code)
+            _write(self._sock, i_pack(len(data) + 4))
+            _write(self._sock, data)
         except ValueError as e:
             if str(e) == "write to closed file":
                 raise InterfaceError("connection is closed")
@@ -757,19 +775,19 @@ class CoreConnection:
         retval.extend(h_pack(0))
 
         self._send_message(BIND, retval)
-        self._write(FLUSH_MSG)
+        _write(self._sock, FLUSH_MSG)
 
     def send_EXECUTE(self):
         """https://www.postgresql.org/docs/current/protocol-message-formats.html"""
-        self._write(EXECUTE_MSG)
-        self._write(FLUSH_MSG)
+        _write(self._sock, EXECUTE_MSG)
+        _write(self._sock, FLUSH_MSG)
 
     def handle_NO_DATA(self, msg, context):
         pass
 
     def handle_COMMAND_COMPLETE(self, data, context):
         if self._transaction_status == IN_FAILED_TRANSACTION and context.error is None:
-            sql = context.statement.strip().rstrip(";").rstrip().upper()
+            sql = context.statement.split()[0].rstrip(";").upper()
             if sql != "ROLLBACK":
                 context.error = InterfaceError("in failed transaction block")
 
@@ -801,13 +819,9 @@ class CoreConnection:
         code = None
 
         while code != READY_FOR_QUERY:
+            code, data_len = ci_unpack(_read(self._sock, 5))
 
-            try:
-                code, data_len = ci_unpack(self._read(5))
-            except struct.error as e:
-                raise InterfaceError("network error") from e
-
-            self.message_types[code](self._read(data_len - 4), context)
+            self.message_types[code](_read(self._sock, data_len - 4), context)
 
         if context.error is not None:
             raise context.error
@@ -815,9 +829,9 @@ class CoreConnection:
     def close_prepared_statement(self, statement_name_bin):
         """https://www.postgresql.org/docs/current/protocol-message-formats.html"""
         self._send_message(CLOSE, STATEMENT + statement_name_bin)
-        self._write(FLUSH_MSG)
-        self._write(SYNC_MSG)
-        self._flush()
+        _write(self._sock, FLUSH_MSG)
+        _write(self._sock, SYNC_MSG)
+        _flush(self._sock)
         context = Context(None)
         self.handle_messages(context)
         self._statement_nums.remove(statement_name_bin)
