@@ -1,4 +1,5 @@
 import datetime
+import socket
 import warnings
 
 import pytest
@@ -25,6 +26,22 @@ def test_internet_socket_connection_refused():
         connect(**conn_params)
 
 
+def test_Connection_plain_socket(db_kwargs):
+    host = db_kwargs.get("host", "localhost")
+    port = db_kwargs.get("port", 5432)
+    with socket.create_connection((host, port)) as sock:
+        user = db_kwargs["user"]
+        password = db_kwargs["password"]
+        conn_params = {"sock": sock, "user": user, "password": password}
+
+        con = connect(**conn_params)
+        cur = con.cursor()
+
+        cur.execute("SELECT 1")
+        res = cur.fetchall()
+        assert res[0][0] == 1
+
+
 def test_database_missing(db_kwargs):
     db_kwargs["database"] = "missing-db"
     with pytest.raises(DatabaseError):
