@@ -23,6 +23,7 @@ from pg8000 import (
     Binary,
     INTERVAL,
     PGInterval,
+    Range,
     pginterval_in,
     pginterval_out,
     timedelta_in,
@@ -334,6 +335,12 @@ def test_int2vector_in(cursor):
     tuple(cursor.execute("select indkey from pg_index"))
 
 
+def test_range_roundtrip(cursor):
+    v = Range(1, 3)
+    retval = tuple(cursor.execute("select cast(%s as int4range)", (v,)))
+    assert retval[0][0] == v
+
+
 def test_timestamp_tz_out(cursor):
     cursor.execute(
         "SELECT '2001-02-03 04:05:06.17 America/Edmonton'" "::timestamp with time zone"
@@ -744,7 +751,7 @@ def test_infinity_timestamp_roundtrip(cursor):
 
 
 def test_point_roundtrip(cursor):
-    v = "(2.3,1)"
+    v = 2.3, 1
     retval = tuple(cursor.execute("SELECT cast(%s as point) as f1", (v,)))
     assert retval[0][0] == v
 
